//
//  GeometryGenerator.cpp
//  LevelSetMethod2
//
//  Created by kumada on 2013/01/19.
//  Copyright (c) 2013年 kumada. All rights reserved.
//

#include "GeometryGenerator.h"
#include <boost/range/algorithm/for_each.hpp>
#include <boost/range/algorithm/transform.hpp>

using namespace lsm;

GeometryGenerator::GeometryGenerator(const SpaceSize3d& space_size)
        : space_size_(space_size) {}

void GeometryGenerator::add_sphere(const Sphere& sphere)
{
        spheres_.push_back(sphere);
}

namespace
{
        struct SphereGenerator
        {
                SphereGenerator(const Sphere& sphere)
                        : sphere_(sphere) {}
                
                const Sphere& sphere_;
                
                void operator()(const IntPoint3d& p, std::uint8_t& v) const
                {
                        int dx = p[0] - sphere_.center_[0];
                        int dy = p[1] - sphere_.center_[1];
                        int dz = p[2] - sphere_.center_[2];
                        double r = std::sqrt(dx * dx + dy * dy + dz * dz);
                        if ( sphere_.is_solid_ ) {
                                if ( r < sphere_.radius_ ) {
                                        v = 128;
                                }
                        } else {
                                if ( r > sphere_.radius_ ) {
                                        v = 128;
                                }
                        }
                }
        };
}

void GeometryGenerator::generate(std::vector<std::uint8_t>& space) const
{
        const int width = space_size_.width_;
        const int height = space_size_.height_;
        const int depth = space_size_.depth_;
        const int area = width * height;
        
        std::vector<SphereGenerator> sphere_generators;
        sphere_generators.reserve(spheres_.size());
        boost::for_each(spheres_, [&](const Sphere& i){ sphere_generators.emplace_back(i); });
        
        for ( int k = 0, ak = area * k; k < depth; ++k, ak += area ) {
                for ( int j = 0, wj = ak + width * j; j < height; ++j, wj += width ) {
                        std::uint8_t* p = &space[wj];
                        for ( int i = 0; i < width; ++i ) {
                                boost::for_each(sphere_generators, [&](const SphereGenerator& sg){ sg({{i, j, k}}, p[i]); });
                        }
                }
        }
}